@@ -21,22 +21,29 @@ class HexSim_simulator(Base_simulator):
         self._phaseStep = 7
         self._angleStep = 1
         super().__init__()
-        self.eta = self.n * np.sqrt(3.0) / 2 / self.NA
+        self._eta = self.n * np.sqrt(3.0) / 2 / self.NA
 
-    def _ill(self):
-        """All polarisations are normalised to average intensity of 1, and with theta being  π/2 for the light sheet"""
+    """All polarisations are normalised to average intensity of 1, and with theta being  π/2 for the light sheet"""
+    def _illCi(self):
         # Circular polarisation
-        self._illCi = 1 + 1 / 6 * (np.cos(self.ph * (-np.sqrt(3) * self.x - self.y) / 2 + self.p2) \
+        _illCi = 1 + 1 / 6 * (np.cos(self.ph * (-np.sqrt(3) * self.x - self.y) / 2 + self.p2) \
                       + np.cos(self.ph * (-np.sqrt(3) * self.x + self.y) / 2 + self.p1) \
                       + np.cos(self.ph * self.y + self.p1 - self.p2))
+        return [_illCi]
+
+    def _illAx(self):
         # Axial polarisation
-        self._illAx = 1 + 2 / 3 * (np.cos(self.ph * self.y + self.p1 - self.p2) \
+        _illAx = 1 + 2 / 3 * (np.cos(self.ph * self.y + self.p1 - self.p2) \
                       + np.cos(self.ph * (self.y - np.sqrt(3) * self.x) / 2 + self.p1) \
                       + np.cos(self.ph * (-self.y - np.sqrt(3) * self.x) / 2 + self.p2))
+        return [_illAx]
+
+    def _illIp(self):
         # In-plane polarisation
-        self._illIp = 1 - 1 / 3 * (np.cos(self.ph * self.y + self.p1 - self.p2) \
+        _illIp = 1 - 1 / 3 * (np.cos(self.ph * self.y + self.p1 - self.p2) \
                       + np.cos(self.ph * (self.y - np.sqrt(3) * self.x) / 2 + self.p1) \
                       + np.cos(self.ph * (self.y + np.sqrt(3) * self.x) / 2 + self.p2))
+        return [_illIp]
 
 class RightHexSim_simulator(Base_simulator):
     '''
@@ -47,18 +54,25 @@ class RightHexSim_simulator(Base_simulator):
         self._phaseStep = 7
         self._angleStep = 1
         super().__init__()
-        self.eta = self.n / self.NA
+        self._eta = self.n * np.sqrt(3.0) / 2 / self.NA
 
-    def _ill(self):
+    def _illCi(self):
+        # Circular polarisation
+        _illCi = 1 + 1 / 3 * (np.cos(self.ph * (-self.x + self.y) + self.p1) +
+                                   np.cos(self.ph * (self.x + self.y) + self.p1 - self.p2))
+        return [_illCi]
+
+    def _illAx(self):
         # Axial polarisation with theta being π/2
-        self._illAx = 1 + 2 / 3 * (np.cos(self.ph * (self.y) + self.p1 - self.p2) +
+        _illAx = 1 + 2 / 3 * (np.cos(self.ph * self.y + self.p1 - self.p2) +
                                    np.cos(self.ph * (self.x + self.y) / 2 + self.p1) +
                                    np.cos(self.ph * (-self.x + self.y) / 2 + self.p2))
+        return [_illAx]
+
+    def _illIp(self):
         # In-plane polarisation
-        self._illIp = 1 - 2 / 3 * np.cos(self.ph * (-self.x) + self.p2)
-        # Circular polarisation
-        self._illCi = 1 + 1 / 3 * (np.cos(self.ph * (-self.x + self.y) + self.p1) +
-                                    np.cos(self.ph * (self.x + self.y) + self.p1 - self.p2))
+        _illIp = 1 - 2 / 3 * np.cos(self.ph * (-self.x) + self.p2)
+        return [_illIp]
 
 
 if __name__ == '__main__':
